@@ -33,15 +33,16 @@ else
 fi
 ####
 # true if we are using the svn version of dzen2
-dzen2_version=$(dzen2 -v 2>&1 | head -n 1 | cut -d , -f 1|cut -d - -f 2)
-if [ -z "$dzen2_version" ] ; then
+# depending on version/distribution, this seems to have version strings like
+# "dzen-" or "dzen-x.x.x-svn"
+if dzen2 -v 2>&1 | head -n 1 | grep -q '^dzen-\([^,]*-svn\|\),'; then
     dzen2_svn="true"
 else
     dzen2_svn=""
 fi
 
 function uniq_linebuffered() {
-    awk '$0 != l { print ; l=$0 ; fflush(); }' "$@"
+    awk -W interactive '$0 != l { print ; l=$0 ; fflush(); }' "$@"
 }
 
 herbstclient pad $monitor $panel_height
@@ -103,11 +104,11 @@ herbstclient pad $monitor $panel_height
         # find out event origin
         case "${cmd[0]}" in
             tag*)
-                #echo "reseting tags" >&2
+                #echo "resetting tags" >&2
                 TAGS=( $(herbstclient tag_status $monitor) )
                 ;;
             date)
-                #echo "reseting date" >&2
+                #echo "resetting date" >&2
                 date="${cmd[@]:1}"
                 ;;
             quit_panel)
