@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -14,8 +14,9 @@
 #include "ewmh.h"
 #include "rules.h"
 #include "ipc-protocol.h"
+#include "object.h"
 // system
-#include <glib.h>
+#include "glib-backports.h"
 #include <assert.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -41,15 +42,20 @@ unsigned long g_window_border_normal_color;
 unsigned long g_window_border_urgent_color;
 unsigned long g_window_border_inner_color;
 
-GHashTable* g_clients; // container of all clients
+static GHashTable* g_clients; // container of all clients
+static HSObject*   g_client_object;
 
 // atoms from dwm.c
 // default atoms
 enum { WMProtocols, WMDelete, WMState, WMTakeFocus, WMLast };
 static Atom g_wmatom[WMLast];
 
+static void client_set_urgent_force(HSClient* client, bool state);
+
 static HSClient* create_client() {
     HSClient* hc = g_new0(HSClient, 1);
+    hsobject_init(&hc->object);
+    hc->window_str = NULL;
     hc->float_size.width = 100;
     hc->float_size.height = 100;
     hc->title = g_string_new("");
@@ -59,6 +65,7 @@ static HSClient* create_client() {
     hc->pseudotile = false;
     hc->ewmhrequests = true;
     hc->ewmhnotify = true;
+    hc->sizehints = true;
     return hc;
 }
 
@@ -87,6 +94,7 @@ void clientlist_init() {
     g_wmatom[WMState] = XInternAtom(g_display, "WM_STATE", False);
     g_wmatom[WMTakeFocus] = XInternAtom(g_display, "WM_TAKE_FOCUS", False);
     // init actual client list
+    g_client_object = hsobject_create_and_link(hsobject_root(), "clients");
     g_clients = g_hash_table_new_full(g_int_hash, g_int_equal,
                                       NULL, (GDestroyNotify)client_destroy);
 }
@@ -114,6 +122,7 @@ void clientlist_destroy() {
     g_hash_table_foreach(g_clients, client_move_to_floatpos, NULL);
 
     g_hash_table_destroy(g_clients);
+    hsobject_unlink_and_destroy(hsobject_root(), g_client_object);
 }
 
 
@@ -125,6 +134,25 @@ HSClient* get_client_from_window(Window window) {
     return (HSClient*) g_hash_table_lookup(g_clients, &window);
 }
 
+#define CLIENT_UPDATE_ATTR(FUNC,MEMBER) do { \
+        HSClient* client = container_of(attr->value.b, HSClient, MEMBER); \
+        FUNC(client, client->MEMBER); \
+        return NULL; \
+    }   \
+    while (0);
+
+static GString* client_attr_fullscreen(HSAttribute* attr) {
+    CLIENT_UPDATE_ATTR(client_set_fullscreen, fullscreen);
+}
+
+static GString* client_attr_pseudotile(HSAttribute* attr) {
+    CLIENT_UPDATE_ATTR(client_set_pseudotile, pseudotile);
+}
+
+static GString* client_attr_urgent(HSAttribute* attr) {
+    CLIENT_UPDATE_ATTR(client_set_urgent_force, urgent);
+}
+
 HSClient* manage_client(Window win) {
     if (is_herbstluft_window(g_display, win)) {
         // ignore our own window
@@ -170,6 +198,9 @@ HSClient* manage_client(Window win) {
 
     // actually manage it
     g_hash_table_insert(g_clients, &(client->window), client);
+    client->window_str = g_string_sized_new(10);
+    g_string_printf(client->window_str, "0x%lx", win);
+    hsobject_link(g_client_object, &client->object, client->window_str->str);
     // insert to layout
     if (!client->tag) {
         client->tag = m->tag;
@@ -190,6 +221,19 @@ HSClient* manage_client(Window win) {
         frame_focus_window(client->tag->frame, win);
     }
 
+    HSAttribute attributes[] = {
+        ATTRIBUTE_STRING(   "winid",        client->window_str,     ATTR_READ_ONLY),
+        ATTRIBUTE_STRING(   "title",        client->title,          ATTR_READ_ONLY),
+        ATTRIBUTE_BOOL(     "fullscreen",   client->fullscreen,     client_attr_fullscreen),
+        ATTRIBUTE_BOOL(     "pseudotile",   client->pseudotile,     client_attr_pseudotile),
+        ATTRIBUTE_BOOL(     "ewmhrequests", client->ewmhrequests,   ATTR_ACCEPT_ALL),
+        ATTRIBUTE_BOOL(     "ewmhnotify",   client->ewmhnotify,     ATTR_ACCEPT_ALL),
+        ATTRIBUTE_BOOL(     "sizehints",    client->sizehints,      ATTR_ACCEPT_ALL),
+        ATTRIBUTE_BOOL(     "urgent",       client->urgent,         client_attr_urgent),
+        ATTRIBUTE_LAST,
+    };
+    hsobject_set_attributes(&client->object, attributes);
+
     ewmh_window_update_tag(client->window, client->tag);
     tag_set_flags_dirty();
     client_set_fullscreen(client, changes.fullscreen);
@@ -224,6 +268,9 @@ void unmanage_client(Window win) {
     if (!client) {
         return;
     }
+    if (client->dragged) {
+        mouse_stop_drag();
+    }
     // remove from tag
     frame_remove_window(client->tag->frame, win);
     // ignore events from it
@@ -242,16 +289,21 @@ void unmanage_client(Window win) {
 
 // destroys a special client
 void client_destroy(HSClient* client) {
+    hsobject_unlink(g_client_object, &client->object);
     if (client->tag && client->slice) {
         stack_remove_slice(client->tag->stack, client->slice);
     }
     if (client->slice) {
         slice_destroy(client->slice);
     }
-    if (client) {
+    if (client->title) {
         /* free window title */
         g_string_free(client->title, true);
     }
+    if (client->window_str) {
+        g_string_free(client->window_str, true);
+    }
+    hsobject_free(&client->object);
     g_free(client);
 }
 
@@ -273,6 +325,7 @@ void window_unfocus_last() {
     if (lastfocus) {
         window_unfocus(lastfocus);
     }
+    hsobject_unlink_by_name(g_client_object, "focus");
     // give focus to root window
     XSetInputFocus(g_display, g_root, RevertToPointerRoot, CurrentTime);
     if (lastfocus) {
@@ -300,6 +353,7 @@ void window_focus(Window window) {
          * only emit the hook if the focus *really* changes */
         // unfocus last one
         window_unfocus(lastfocus);
+        hsobject_link(g_client_object, &client->object, "focus");
         ewmh_update_active_window(window);
         tag_update_each_focus_layer();
         char* title = client ? client->title->str : "?";
@@ -402,16 +456,21 @@ void client_resize(HSClient* client, XRectangle rect, HSFrame* frame) {
     // apply border width
     rect.width -= border_width * 2;
     rect.height -= border_width * 2;
-    if (!*g_smart_window_surroundings
-        || (frame->content.clients.count != 1
-            && frame->content.clients.layout != LAYOUT_MAX)) {
+    bool is_max_layout = frame->content.clients.layout != LAYOUT_MAX;
+    bool only_one_client = frame->content.clients.count != 1;
+    bool smart_surroundings_are_applied = *g_smart_window_surroundings && is_max_layout && only_one_client;
+    if (!client->pseudotile && !smart_surroundings_are_applied) {
         // apply window gap
         rect.width -= *g_window_gap;
         rect.height -= *g_window_gap;
     }
 
-    XSetWindowBorderWidth(g_display, win, border_width);
-    XMoveResizeWindow(g_display, win, rect.x, rect.y, rect.width, rect.height);
+    XWindowChanges changes = {
+      .x = rect.x, .y = rect.y, .width = rect.width, .height = rect.height,
+      .border_width = border_width
+    };
+    int mask = CWX | CWY | CWWidth | CWHeight | CWBorderWidth;
+    XConfigureWindow(g_display, win, mask, &changes);
     if (*g_window_border_inner_width > 0
         && *g_window_border_inner_width < *g_window_border_width) {
         unsigned long current_border_color = get_window_border_color(client);
@@ -461,7 +520,11 @@ void client_resize_floating(HSClient* client, HSMonitor* m) {
     if (client->float_size.height < WINDOW_MIN_HEIGHT)
         client->float_size.height = WINDOW_MIN_HEIGHT;
 
-    client->last_border_width = *g_window_border_width;
+    bool border_changed = false;
+    if (client->last_border_width != *g_window_border_width) {
+        client->last_border_width = *g_window_border_width;
+        border_changed = true;
+    }
     client->last_size = client->float_size;
     client->last_size.x += m->rect.x + m->pad_left;
     client->last_size.y += m->rect.y + m->pad_up;
@@ -477,7 +540,12 @@ void client_resize_floating(HSClient* client, HSMonitor* m) {
               m->rect.y + m->pad_up - client->last_size.height + space,
               m->rect.y + m->rect.height - m->pad_up - m->pad_down - space);
     XRectangle rect = client->last_size;
-    XSetWindowBorderWidth(g_display, client->window, *g_window_border_width);
+    // add window border to last_size
+    client->last_size.width += 2 * client->last_border_width;
+    client->last_size.height += 2 * client->last_border_width;
+    if (border_changed) {
+        XSetWindowBorderWidth(g_display, client->window, *g_window_border_width);
+    }
     XMoveResizeWindow(g_display, client->window,
         rect.x, rect.y, rect.width, rect.height);
     if (*g_window_border_inner_width > 0
@@ -546,7 +614,10 @@ void client_set_urgent(HSClient* client, bool state) {
         // nothing to do
         return;
     }
+    client_set_urgent_force(client, state);
+}
 
+void client_set_urgent_force(HSClient* client, bool state) {
     char winid_str[STRING_BUF_SIZE];
     snprintf(winid_str, STRING_BUF_SIZE, "0x%lx", client->window);
     hook_emit_list("urgent", state ? "on" : "off", winid_str, NULL);
@@ -635,11 +706,6 @@ HSClient* get_current_client() {
 }
 
 void client_set_fullscreen(HSClient* client, bool state) {
-    if (client->fullscreen == state) {
-        // nothing to do
-        return;
-    }
-
     client->fullscreen = state;
     if (client->ewmhnotify) {
         client->ewmhfullscreen = state;
@@ -694,8 +760,11 @@ int client_set_property_command(int argc, char** argv) {
     }
 
     // if found, then change it
+    bool old_value = *(properties[i].value);
     bool state = string_to_bool(action, *(properties[i].value));
-    properties[i].func(client, state);
+    if (state != old_value) {
+        properties[i].func(client, state);
+    }
     return 0;
 }
 
