@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -14,7 +14,7 @@
 #include "stack.h"
 #include "monitor.h"
 
-#include <glib.h>
+#include "glib-backports.h"
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
@@ -205,6 +205,25 @@ HSFrame* lookup_frame(HSFrame* root, char *index) {
     return lookup_frame(new_root, new_index);
 }
 
+HSFrame* find_frame_with_window(HSFrame* frame, Window window) {
+    if (frame->type == TYPE_CLIENTS) {
+        Window* buf = frame->content.clients.buf;
+        size_t count = frame->content.clients.count;
+        for (int i = 0; i < count; i++) {
+            if (buf[i] == window) {
+                return frame;
+            }
+        }
+        return NULL;
+    } else { /* frame->type == TYPE_FRAMES */
+        HSFrame* found = find_frame_with_window(frame->content.layout.a, window);
+        if (!found) {
+            found = find_frame_with_window(frame->content.layout.b, window);
+        }
+        return found;
+    }
+}
+
 bool frame_remove_window(HSFrame* frame, Window window) {
     if (frame->type == TYPE_CLIENTS) {
         Window* buf = frame->content.clients.buf;
@@ -878,10 +897,7 @@ void frame_update_frame_window_visibility(HSFrame* frame) {
     frame_do_recursive(frame, frame_update_frame_window_visibility_helper, 2);
 }
 
-HSFrame* frame_current_selection() {
-    HSMonitor* m = get_current_monitor();
-    if (!m->tag) return NULL;
-    HSFrame* frame = m->tag->frame;
+HSFrame* frame_current_selection_below(HSFrame* frame) {
     while (frame->type == TYPE_FRAMES) {
         frame = (frame->content.layout.selection == 0) ?
                 frame->content.layout.a :
@@ -890,6 +906,12 @@ HSFrame* frame_current_selection() {
     return frame;
 }
 
+HSFrame* frame_current_selection() {
+    HSMonitor* m = get_current_monitor();
+    if (!m->tag) return NULL;
+    return frame_current_selection_below(m->tag->frame);
+}
+
 int frame_current_bring(int argc, char** argv, GString* output) {
     HSClient* client = NULL;
 
@@ -1114,6 +1136,13 @@ int cycle_all_command(int argc, char** argv) {
         index %= frame->content.clients.count;
         frame->content.clients.selection = index;
     }
+    Window w = frame_focused_window(g_cur_frame);
+    if (w) {
+        HSClient* c = get_client_from_window(w);
+        if (c) {
+            client_raise(c);
+        }
+    }
     monitor_apply_layout(get_current_monitor());
     return 0;
 }
@@ -1472,6 +1501,8 @@ Window frame_focused_window(HSFrame* frame) {
 }
 
 // try to focus window in frame
+// it does not require anything from the frame. it may be infocused or even
+// hidden.
 // returns true if win was found and focused, else returns false
 bool frame_focus_window(HSFrame* frame, Window win) {
     if (!frame) {
@@ -1599,6 +1630,22 @@ void frame_do_recursive(HSFrame* frame, void (*action)(HSFrame*), int order) {
     }
 }
 
+void frame_do_recursive_data(HSFrame* frame, void (*action)(HSFrame*,void*),
+                             int order, void* data) {
+    if (frame->type == TYPE_FRAMES) {
+        // clients and subframes
+        HSLayout* layout = &(frame->content.layout);
+        if (order <= 0) action(frame, data);
+        frame_do_recursive_data(layout->a, action, order, data);
+        if (order == 1) action(frame, data);
+        frame_do_recursive_data(layout->b, action, order, data);
+        if (order >= 2) action(frame, data);
+    } else {
+        // action only
+        action(frame, data);
+    }
+}
+
 static void frame_hide(HSFrame* frame) {
     frame_set_visible(frame, false);
     if (frame->type == TYPE_CLIENTS) {
