@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -18,6 +18,7 @@
 #include "rules.h"
 #include "ewmh.h"
 #include "stack.h"
+#include "object.h"
 // standard
 #include <string.h>
 #include <stdio.h>
@@ -51,6 +52,9 @@ typedef void (*HandlerTable[LASTEvent]) (XEvent*);
 int quit();
 int reload();
 int version(int argc, char* argv[], GString* output);
+int echo(int argc, char* argv[], GString* output);
+int true_command();
+int false_command();
 int print_layout_command(int argc, char** argv, GString* output);
 int load_command(int argc, char** argv, GString* output);
 int print_tag_status_command(int argc, char** argv, GString* output);
@@ -85,6 +89,9 @@ void unmapnotify(XEvent* event);
 
 CommandBinding g_commands[] = {
     CMD_BIND_NO_OUTPUT(   "quit",           quit),
+    CMD_BIND(             "echo",           echo),
+    CMD_BIND_NO_OUTPUT(   "true",           true_command),
+    CMD_BIND_NO_OUTPUT(   "false",          false_command),
     CMD_BIND_NO_OUTPUT(   "reload",         reload),
     CMD_BIND(             "version",        version),
     CMD_BIND(             "list_commands",  list_commands),
@@ -145,11 +152,13 @@ CommandBinding g_commands[] = {
     CMD_BIND(             "raise",          raise_command),
     CMD_BIND(             "rule",           rule_add_command),
     CMD_BIND(             "unrule",         rule_remove_command),
+    CMD_BIND(             "list_rules",     rule_print_all_command),
     CMD_BIND(             "layout",         print_layout_command),
     CMD_BIND(             "stack",          print_stack_command),
     CMD_BIND(             "dump",           print_layout_command),
     CMD_BIND(             "load",           load_command),
     CMD_BIND(             "complete",       complete_command),
+    CMD_BIND(             "complete_shell", complete_command),
     CMD_BIND_NO_OUTPUT(   "lock",           monitors_lock_command),
     CMD_BIND_NO_OUTPUT(   "unlock",         monitors_unlock_command),
     CMD_BIND(             "lock_tag",       monitor_lock_tag_command),
@@ -160,6 +169,16 @@ CommandBinding g_commands[] = {
     CMD_BIND(             "and",            command_chain_command),
     CMD_BIND(             "or",             command_chain_command),
     CMD_BIND(             "!",              negate_command),
+    CMD_BIND(             "attr",           attr_command),
+    CMD_BIND(             "compare",        compare_command),
+    CMD_BIND(             "object_tree",    print_object_tree_command),
+    CMD_BIND(             "get_attr",       hsattribute_get_command),
+    CMD_BIND(             "set_attr",       hsattribute_set_command),
+    CMD_BIND(             "new_attr",       userattribute_command),
+    CMD_BIND(             "mktemp",         tmpattribute_command),
+    CMD_BIND(             "remove_attr",    userattribute_remove_command),
+    CMD_BIND(             "substitute",     substitute_command),
+    CMD_BIND(             "sprintf",        sprintf_command),
     CMD_BIND(             "getenv",         getenv_command),
     CMD_BIND(             "setenv",         setenv_command),
     CMD_BIND(             "unsetenv",       unsetenv_command),
@@ -185,6 +204,28 @@ int version(int argc, char* argv[], GString* output) {
     return 0;
 }
 
+int echo(int argc, char* argv[], GString* output) {
+    if (SHIFT(argc, argv)) {
+        // if there still is an argument
+        g_string_append(output, argv[0]);
+        while (SHIFT(argc, argv)) {
+            g_string_append_c(output, ' ');
+            g_string_append(output, argv[0]);
+        }
+    }
+    g_string_append_c(output, '\n');
+    return 0;
+}
+
+int true_command() {
+    return 0;
+}
+
+int false_command() {
+    return 1;
+}
+
+
 // prints or dumps the layout of an given tag
 // first argument tells whether to print or to dump
 int print_layout_command(int argc, char** argv, GString* output) {
@@ -442,18 +483,25 @@ void event_on_configure(XEvent event) {
     ce.event = cre->window;
     ce.window = cre->window;
     if (client) {
-        cre->width += 2*cre->border_width - 2*client->last_border_width;
-        cre->height += 2*cre->border_width - 2*client->last_border_width;
-        client->float_size.width = cre->width;
-        client->float_size.height = cre->height;
-        ce.x = client->last_size.x;
-        ce.y = client->last_size.y;
-        ce.width = client->last_size.width;
-        ce.height = client->last_size.height;
-        ce.override_redirect = False;
-        ce.border_width = cre->border_width;
-        ce.above = cre->above;
-        if (client->tag->floating || client->pseudotile) {
+        bool changes = false;
+        if (client->sizehints && !client->dragged) {
+            cre->width += 2*cre->border_width - 2*client->last_border_width;
+            cre->height += 2*cre->border_width - 2*client->last_border_width;
+            if (client->float_size.width != cre->width) changes = true;
+            if (client->float_size.height != cre->height) changes = true;
+            client->float_size.width = cre->width;
+            client->float_size.height = cre->height;
+            ce.x = client->last_size.x;
+            ce.y = client->last_size.y;
+            ce.width = client->last_size.width;
+            ce.height = client->last_size.height;
+            ce.override_redirect = False;
+            ce.border_width = cre->border_width;
+            ce.above = cre->above;
+        }
+        if (changes && client->tag->floating) {
+            client_resize_floating(client, find_monitor_with_tag(client->tag));
+        } else if (changes && client->pseudotile) {
             monitor_apply_layout(find_monitor_with_tag(client->tag));
         } else {
         // FIXME: why send event and not XConfigureWindow or XMoveResizeWindow??
@@ -617,7 +665,7 @@ static void parse_arguments(int argc, char** argv) {
                 break;
             case 'v':
                 printf("%s %s\n", argv[0], HERBSTLUFT_VERSION);
-                printf("Copyright (c) 2011-2012 Thorsten Wißmann\n");
+                printf("Copyright (c) 2011-2013 Thorsten Wißmann\n");
                 printf("Released under the Simplified BSD License\n");
                 exit(0);
                 break;
@@ -683,6 +731,7 @@ static struct {
     void (*destroy)();
 } g_modules[] = {
     { ipc_init,         ipc_destroy         },
+    { object_tree_init, object_tree_destroy },
     { key_init,         key_destroy         },
     { settings_init,    settings_destroy    },
     { stacklist_init,   stacklist_destroy   },
@@ -754,8 +803,17 @@ void enternotify(XEvent* event) {
     if (!mouse_is_dragging()
         && *g_focus_follows_mouse
         && false == ce->focus) {
-        // sloppy focus
-        focus_window(ce->window, false, true);
+        HSClient* c = get_client_from_window(ce->window);
+        HSFrame* target;
+        if (c && c->tag->floating == false
+              && (target = find_frame_with_window(c->tag->frame, ce->window))
+              && target->content.clients.layout == LAYOUT_MAX
+              && frame_focused_window(target) != ce->window) {
+            // don't allow focus_follows_mouse if another window would be
+            // hidden during that focus change (which only occurs in max layout)
+        } else {
+            focus_window(ce->window, false, true);
+        }
     }
 }
 
