@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -10,8 +10,9 @@
 #include "clientlist.h"
 #include "ipc-protocol.h"
 #include "hook.h"
+#include "command.h"
 
-#include <glib.h>
+#include "glib-backports.h"
 #include "glib-backports.h"
 #include <string.h>
 #include <stdio.h>
@@ -75,8 +76,9 @@ static HSConditionType g_condition_types[] = {
     { "windowrole",     condition_windowrole        },
 };
 
-int     g_maxage_type; // index of "maxage"
-time_t  g_current_rule_birth_time; // data from rules_apply() to condition_maxage()
+static int     g_maxage_type; // index of "maxage"
+static time_t  g_current_rule_birth_time; // data from rules_apply() to condition_maxage()
+static unsigned long long g_rule_label_index; // incremental index of rule label
 
 static HSConsequenceType g_consequence_types[] = {
     { "tag",            consequence_tag             },
@@ -91,12 +93,13 @@ static HSConsequenceType g_consequence_types[] = {
     { "hook",           consequence_hook            },
 };
 
-GQueue g_rules = G_QUEUE_INIT; // a list of HSRule* elements
+static GQueue g_rules = G_QUEUE_INIT; // a list of HSRule* elements
 
 /// FUNCTIONS ///
 // RULES //
 void rules_init() {
     g_maxage_type = find_condition_type("maxage");
+    g_rule_label_index = 0;
 }
 
 void rules_destroy() {
@@ -141,15 +144,16 @@ HSCondition* condition_create(int type, char op, char* value, GString* output) {
 
         case '~':
             cond.value_type = CONDITION_VALUE_TYPE_REGEX;
-            int status = regcomp(&cond.value.exp, value, REG_EXTENDED);
+            int status = regcomp(&cond.value.reg.exp, value, REG_EXTENDED);
             if (status != 0) {
                 char buf[ERROR_STRING_BUF_SIZE];
-                regerror(status, &cond.value.exp, buf, ERROR_STRING_BUF_SIZE);
+                regerror(status, &cond.value.reg.exp, buf, ERROR_STRING_BUF_SIZE);
                 g_string_append_printf(output,
                     "rule: Can not parse value \"%s\" from condition \"%s\": \"%s\"\n",
                     value, g_condition_types[type].name, buf);
                 return NULL;
             }
+            cond.value.reg.str = g_strdup(value);
             break;
 
         default:
@@ -176,7 +180,8 @@ void condition_destroy(HSCondition* cond) {
             free(cond->value.str);
             break;
         case CONDITION_VALUE_TYPE_REGEX:
-            regfree(&cond->value.exp);
+            regfree(&cond->value.reg.exp);
+            g_free(cond->value.reg.str);
             break;
         default:
             break;
@@ -230,12 +235,35 @@ void consequence_destroy(HSConsequence* cons) {
     g_free(cons);
 }
 
+bool rule_label_replace(HSRule* rule, char op, char* value, GString* output) {
+    switch (op) {
+        case '=':
+            if (*value == '\0') {
+                g_string_append_printf(output,
+                    "rule: Rule label cannot be empty");
+                return false;
+                break;
+            }
+            g_free(rule->label);
+            rule->label = g_strdup(value);
+            break;
+        default:
+            g_string_append_printf(output,
+                "rule: Unknown rule label operation \"%c\"\n", op);
+            return false;
+            break;
+    }
+    return true;
+}
+
 // rules parsing //
 
 HSRule* rule_create() {
     HSRule* rule = g_new0(HSRule, 1);
     rule->once = false;
     rule->birth_time = get_monotonic_timestamp();
+    // Add name. Defaults to index number.
+    rule->label = g_strdup_printf("%llu", g_rule_label_index++);
     return rule;
 }
 
@@ -250,10 +278,106 @@ void rule_destroy(HSRule* rule) {
         consequence_destroy(rule->consequences[i]);
     }
     g_free(rule->consequences);
+    // free label
+    g_free(rule->label);
     // free rule itself
     g_free(rule);
 }
 
+void rule_complete(int argc, char** argv, int pos, GString* output) {
+    char* needle = (pos < argc) ? argv[pos] : "";
+    GString* buf = g_string_sized_new(20);
+
+    // complete against conditions
+    for (int i = 0; i < LENGTH(g_condition_types); i++) {
+        g_string_printf(buf, "%s=", g_condition_types[i].name);
+        try_complete_partial(needle, buf->str, output);
+        g_string_printf(buf, "%s~", g_condition_types[i].name);
+        try_complete_partial(needle, buf->str, output);
+    }
+
+    // complete against consequences
+    for (int i = 0; i < LENGTH(g_consequence_types); i++) {
+        g_string_printf(buf, "%s=", g_consequence_types[i].name);
+        try_complete_partial(needle, buf->str, output);
+    }
+
+    // complete label
+    try_complete_partial(needle, "label=", output);
+    // complete flags
+    try_complete(needle, "once",    output);
+    try_complete(needle, "not",     output);
+    try_complete(needle, "!",       output);
+
+    g_string_free(buf, true);
+}
+
+// Compares the id of two rules.
+static gint rule_compare_label(const HSRule* a, const HSRule* b) {
+    return strcmp(a->label, b->label);
+}
+
+// Looks up rules of a given label and removes them from the queue
+bool rule_find_pop(char* label) {
+    GList* rule = { NULL };
+    bool status = false; // Will be returned as true if any is found
+    HSRule rule_find = { .label = label };
+    while ((rule = g_queue_find_custom(&g_rules, &rule_find,
+                        (GCompareFunc)rule_compare_label))) {
+        // Check if rule with label exists
+        if ( rule == NULL ) {
+            break;
+        }
+        status = true;
+        // If so, clear data
+        rule_destroy(rule->data);
+        // Remove and free empty link
+        g_queue_delete_link(&g_rules, rule);
+    }
+    return status;
+}
+
+// List all rules in queue
+static void rule_print_append_output(HSRule* rule, GString* output) {
+    g_string_append_printf(output, "label=%s\t", rule->label);
+    // Append conditions
+    for (int i = 0; i < rule->condition_count; i++) {
+        if (rule->conditions[i]->negated) { // Include flag if negated
+            g_string_append_printf(output, "not\t");
+        }
+        g_string_append_printf(output, "%s=",
+            g_condition_types[rule->conditions[i]->condition_type].name);
+        switch (rule->conditions[i]->value_type) {
+            case CONDITION_VALUE_TYPE_STRING:
+                g_string_append_printf(output, "%s\t",
+                    rule->conditions[i]->value.str);
+                break;
+            case CONDITION_VALUE_TYPE_REGEX:
+                g_string_append_printf(output, "%s\t",
+                    rule->conditions[i]->value.reg.str);
+                break;
+            default: /* CONDITION_VALUE_TYPE_INTEGER: */
+                g_string_append_printf(output, "%i\t",
+                    rule->conditions[i]->value.integer);
+                break;
+        }
+    }
+    // Append consequences
+    for (int i = 0; i < rule->consequence_count; i++) {
+        g_string_append_printf(output, "%s=%s\t",
+            g_consequence_types[rule->consequences[i]->type].name,
+            rule->consequences[i]->value.str);
+    }
+    // Print new line
+    g_string_append_c(output, '\n');
+}
+
+int rule_print_all_command(int argc, char** argv, GString* output) {
+    // Print entry for each in the queue
+    g_queue_foreach(&g_rules, (GFunc)rule_print_append_output, output);
+    return 0;
+}
+
 // parses an arg like NAME=VALUE to res_name, res_operation and res_value
 bool tokenize_arg(char* condition,
                   char** res_name, char* res_operation, char** res_value) {
@@ -304,6 +428,7 @@ int rule_add_command(int argc, char** argv, GString* output) {
     HSRule* rule = rule_create();
     HSCondition* cond;
     HSConsequence* cons;
+    bool printlabel = false;
     bool negated = false;
     struct {
         char* name;
@@ -312,6 +437,7 @@ int rule_add_command(int argc, char** argv, GString* output) {
         { "not",    &negated },
         { "!",      &negated },
         { "once",   &rule->once },
+        { "printlabel",&printlabel },
     };
 
     // parse rule incrementally. always maintain a correct rule in rule
@@ -358,6 +484,14 @@ int rule_add_command(int argc, char** argv, GString* output) {
             rule_add_consequence(rule, cons);
         }
 
+        // Check for a provided label, and replace default index if so
+        else if (consorcond && (!strcmp(name,"label"))) {
+            if (!rule_label_replace(rule, op, value, output)) {
+                rule_destroy(rule);
+                return HERBST_INVALID_ARGUMENT;
+            }
+        }
+
         else {
             // need to hardcode "rule:" here because args are shifted
             g_string_append_printf(output,
@@ -367,10 +501,30 @@ int rule_add_command(int argc, char** argv, GString* output) {
         }
     }
 
+    if (printlabel) {
+       g_string_append_printf(output, "%s\n", rule->label);
+    }
+
     g_queue_push_tail(&g_rules, rule);
     return 0;
 }
 
+void complete_against_rule_names(int argc, char** argv, int pos, GString* output) {
+    char* needle;
+    if (pos >= argc) {
+        needle = "";
+    } else {
+        needle = argv[pos];
+    }
+    // Complete labels
+    GList* cur_rule = g_queue_peek_head_link(&g_rules);
+    while (cur_rule != NULL) {
+        try_complete(needle, ((HSRule*)cur_rule->data)->label, output);
+        cur_rule = g_list_next(cur_rule);
+    }
+}
+
+
 int rule_remove_command(int argc, char** argv, GString* output) {
     if (argc < 2) {
         return HERBST_NEED_MORE_ARGS;
@@ -380,12 +534,16 @@ int rule_remove_command(int argc, char** argv, GString* output) {
         // remove all rules
         g_queue_foreach(&g_rules, (GFunc)rule_destroy, NULL);
         g_queue_clear(&g_rules);
+        g_rule_label_index = 0;
         return 0;
     }
 
-    g_string_append_printf(output,
-        "%s: This command must be used with --all/-F\n", argv[0]);
-    return HERBST_INVALID_ARGUMENT;
+    // Deletes rule with given label
+    if (!rule_find_pop(argv[1])) {
+        g_string_append_printf(output, "Couldn't find rule: \"%s\"", argv[1]);
+        return HERBST_INVALID_ARGUMENT;
+    }
+    return 0;
 }
 
 // rules applying //
@@ -482,7 +640,7 @@ bool condition_string(HSCondition* rule, char* string) {
             return !strcmp(string, rule->value.str);
             break;
         case CONDITION_VALUE_TYPE_REGEX:
-            status = regexec(&rule->value.exp, string, 1, &match, 0);
+            status = regexec(&rule->value.reg.exp, string, 1, &match, 0);
             // only accept it, if it matches the entire string
             if (status == 0
                 && match.rm_so == 0
@@ -651,6 +809,10 @@ void consequence_ewmhnotify(HSConsequence* cons, HSClient* client,
 
 void consequence_hook(HSConsequence* cons, HSClient* client,
                             HSClientChanges* changes) {
-    char* hook_str[] = { "rule" , cons->value.str };
+    GString* winid = g_string_sized_new(20);
+    g_string_printf(winid, "0x%lx", client->window);
+    char* hook_str[] = { "rule" , cons->value.str, winid->str };
     hook_emit(LENGTH(hook_str), hook_str);
+    g_string_free(winid, true);
 }
+
