@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -7,7 +7,7 @@
 #define __HERBST_STACK_H_
 
 #include <X11/Xlib.h>
-#include <glib.h>
+#include "glib-backports.h"
 #include <stdbool.h>
 
 typedef enum Layer {
