@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -21,12 +21,23 @@
 #include "settings.h"
 
 bool    g_tag_flags_dirty = true;
+HSObject* g_tag_object;
+HSObject* g_tag_by_name;
 int* g_raise_on_focus_temporarily;
 
+static int tag_rename(HSTag* tag, char* name, GString* output);
+
 void tag_init() {
     g_tags = g_array_new(false, false, sizeof(HSTag*));
     g_raise_on_focus_temporarily = &(settings_find("raise_on_focus_temporarily")
                                      ->value.i);
+    g_tag_object = hsobject_create_and_link(hsobject_root(), "tags");
+    HSAttribute attributes[] = {
+        ATTRIBUTE_UINT("count", g_tags->len, ATTR_READ_ONLY),
+        ATTRIBUTE_LAST,
+    };
+    hsobject_set_attributes(g_tag_object, attributes);
+    g_tag_by_name = hsobject_create_and_link(g_tag_object, "by-name");
 }
 
 static void tag_free(HSTag* tag) {
@@ -39,7 +50,9 @@ static void tag_free(HSTag* tag) {
         }
     }
     stack_destroy(tag->stack);
+    hsobject_unlink_and_destroy(g_tag_by_name, tag->object);
     g_string_free(tag->name, true);
+    g_string_free(tag->display_name, true);
     g_free(tag);
 }
 
@@ -51,6 +64,8 @@ void tag_destroy() {
         tag_free(tag);
     }
     g_array_free(g_tags, true);
+    hsobject_unlink_and_destroy(g_tag_object, g_tag_by_name);
+    hsobject_unlink_and_destroy(hsobject_root(), g_tag_object);
 }
 
 
@@ -122,6 +137,71 @@ HSTag* find_unused_tag() {
     return NULL;
 }
 
+static GString* tag_attr_floating(HSAttribute* attr) {
+    HSTag* tag = container_of(attr->value.b, HSTag, floating);
+    HSMonitor* m = find_monitor_with_tag(tag);
+    if (m != NULL) {
+        monitor_apply_layout(m);
+    }
+    return NULL;
+}
+
+static GString* tag_attr_name(HSAttribute* attr) {
+    HSTag* tag = container_of(attr->value.str, HSTag, display_name);
+    GString* error = g_string_new("");
+    int status = tag_rename(tag, tag->display_name->str, error);
+    if (status == 0) {
+        g_string_free(error, true);
+        return NULL;
+    } else {
+        return error;
+    }
+}
+
+static void sum_up_clientframes(HSFrame* frame, void* data) {
+    if (frame->type == TYPE_CLIENTS) {
+        (*(int*)data)++;
+    }
+}
+
+static int tag_attr_frame_count(void* data) {
+    HSTag* tag = (HSTag*) data;
+    int i = 0;
+    frame_do_recursive_data(tag->frame, sum_up_clientframes, 0, &i);
+    return i;
+}
+
+static void sum_up_clients(HSFrame* frame, void* data) {
+    if (frame->type == TYPE_CLIENTS) {
+        *(int*)data += frame->content.clients.count;
+    }
+}
+
+static int tag_attr_client_count(void* data) {
+    HSTag* tag = (HSTag*) data;
+    int i = 0;
+    frame_do_recursive_data(tag->frame, sum_up_clients, 0, &i);
+    return i;
+}
+
+
+static int tag_attr_curframe_windex(void* data) {
+    HSTag* tag = (HSTag*) data;
+    HSFrame* frame = frame_current_selection_below(tag->frame);
+    return frame->content.clients.selection;
+}
+
+static int tag_attr_curframe_wcount(void* data) {
+    HSTag* tag = (HSTag*) data;
+    HSFrame* frame = frame_current_selection_below(tag->frame);
+    return frame->content.clients.count;
+}
+
+static int tag_attr_index(void* data) {
+    HSTag* tag = (HSTag*) data;
+    return tag_index_of(tag);
+}
+
 HSTag* add_tag(char* name) {
     HSTag* find_result = find_tag(name);
     if (find_result) {
@@ -132,8 +212,25 @@ HSTag* add_tag(char* name) {
     tag->stack = stack_create();
     tag->frame = frame_create_empty(NULL, tag);
     tag->name = g_string_new(name);
+    tag->display_name = g_string_new(name);
     tag->floating = false;
     g_array_append_val(g_tags, tag);
+
+    // create object
+    tag->object = hsobject_create_and_link(g_tag_by_name, name);
+    tag->object->data = tag;
+    HSAttribute attributes[] = {
+        ATTRIBUTE_STRING("name",           tag->display_name,        tag_attr_name),
+        ATTRIBUTE_BOOL(  "floating",       tag->floating,            tag_attr_floating),
+        ATTRIBUTE_CUSTOM_INT("index",          tag_attr_index,           ATTR_READ_ONLY),
+        ATTRIBUTE_CUSTOM_INT("frame_count",    tag_attr_frame_count,     ATTR_READ_ONLY),
+        ATTRIBUTE_CUSTOM_INT("client_count",   tag_attr_client_count,    ATTR_READ_ONLY),
+        ATTRIBUTE_CUSTOM_INT("curframe_windex",tag_attr_curframe_windex, ATTR_READ_ONLY),
+        ATTRIBUTE_CUSTOM_INT("curframe_wcount",tag_attr_curframe_wcount, ATTR_READ_ONLY),
+        ATTRIBUTE_LAST,
+    };
+    hsobject_set_attributes(tag->object, attributes);
+
     ewmh_update_desktops();
     ewmh_update_desktop_names();
     tag_set_flags_dirty();
@@ -154,6 +251,20 @@ int tag_add_command(int argc, char** argv, GString* output) {
     return 0;
 }
 
+static int tag_rename(HSTag* tag, char* name, GString* output) {
+    if (find_tag(name)) {
+        g_string_append_printf(output,
+            "Error: Tag \"%s\" already exists\n", name);
+        return HERBST_TAG_IN_USE;
+    }
+    hsobject_link_rename(g_tag_by_name, tag->name->str, name);
+    g_string_assign(tag->name, name);
+    g_string_assign(tag->display_name, name);
+    ewmh_update_desktop_names();
+    hook_emit_list("tag_renamed", tag->name->str, NULL);
+    return 0;
+}
+
 int tag_rename_command(int argc, char** argv, GString* output) {
     if (argc < 3) {
         return HERBST_NEED_MORE_ARGS;
@@ -164,15 +275,7 @@ int tag_rename_command(int argc, char** argv, GString* output) {
             "%s: Tag \"%s\" not found\n", argv[0], argv[1]);
         return HERBST_INVALID_ARGUMENT;
     }
-    if (find_tag(argv[2])) {
-        g_string_append_printf(output,
-            "%s: Tag \"%s\" already exists\n", argv[0], argv[2]);
-        return HERBST_TAG_IN_USE;
-    }
-    g_string_assign(tag->name, argv[2]);
-    ewmh_update_desktop_names();
-    hook_emit_list("tag_renamed", tag->name->str, NULL);
-    return 0;
+    return tag_rename(tag, argv[2], output);
 }
 
 int tag_remove_command(int argc, char** argv, GString* output) {
@@ -259,10 +362,7 @@ int tag_set_floating_command(int argc, char** argv, GString* output) {
         }
     }
 
-    bool new_value = false;
-    if (!strcmp(action, "toggle"))      new_value = ! tag->floating;
-    else if (!strcmp(action, "on"))     new_value = true;
-    else if (!strcmp(action, "off"))    new_value = false;
+    bool new_value = string_to_bool(action, tag->floating);
 
     if (!strcmp(action, "status")) {
         // just print status
@@ -386,6 +486,8 @@ void tag_move_client(HSClient* client, HSTag* target) {
     frame_remove_window(tag_source->frame, client->window);
     // insert window into target
     frame_insert_window(target->frame, client->window);
+    // enfoce it to be focused on the target tag
+    frame_focus_window(target->frame, client->window);
     stack_remove_slice(client->tag->stack, client->slice);
     client->tag = target;
     stack_insert_slice(client->tag->stack, client->slice);
@@ -444,3 +546,7 @@ void tag_update_each_focus_layer() {
     tag_foreach(tag_update_focus_layer_helper, NULL);
 }
 
+void tag_update_focus_objects() {
+    hsobject_link(g_tag_object, get_current_monitor()->tag->object, "focus");
+}
+
