@@ -1,4 +1,4 @@
-/** Copyright 2011-2012 Thorsten Wißmann. All rights reserved.
+/** Copyright 2011-2013 Thorsten Wißmann. All rights reserved.
  *
  * This software is licensed under the "Simplified BSD License".
  * See LICENSE for details */
@@ -6,7 +6,7 @@
 #ifndef __HERBST_UTILS_H_
 #define __HERBST_UTILS_H_
 
-#include <glib.h>
+#include "glib-backports.h"
 #include <stddef.h>
 #include <stdbool.h>
 #include <X11/Xlib.h>
@@ -16,6 +16,10 @@
 #define SHIFT(ARGC, ARGV) (--(ARGC) && ++(ARGV))
 #define MOD(X, N) ((((X) % (signed)(N)) + (signed)(N)) % (signed)(N))
 
+#define container_of(ptr, type, member) \
+    ((type *)( (char *)(ptr)- offsetof(type,member) ))
+
+
 /// print a printf-like message to stderr and exit
 void die(const char *errstr, ...);
 
@@ -49,6 +53,7 @@ bool is_window_mapped(Display* dpy, Window window);
 
 bool window_has_property(Display* dpy, Window window, char* prop_name);
 
+bool string_to_bool_error(char* string, bool oldvalue, bool* error);
 bool string_to_bool(char* string, bool oldvalue);
 
 char* strlasttoken(char* str, char* delim);
@@ -107,5 +112,11 @@ void set_window_double_border(Display *dpy, Window win, int ibw,
         (a).height == (b).height    \
     )
 
+// returns an posix sh escaped string or NULL if there is nothing to escape
+// if a new string is returned, then the caller has to free it
+char* posix_sh_escape(char* source);
+// does the reverse action to posix_sh_escape by modifing the string
+void posix_sh_compress_inplace(char* str);
+
 #endif
 
